/*
Modified GeckoTIP code is Copyright 2005-2006, Ian Weiner
---
Original GPL'd code from TPCTIP extension is Copyright (c) 2004, Datagrove, Inc. All rights reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Datagrove, Inc. nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


//#include <windowsx.h>  // DEBUG

#define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers
#include <windows.h>
#include <winuser.h>
#include <oleacc.h>
#include "InputScope.h"
#include <stdio.h>

#define MOZILLA_STRICT_API
#define XP_WIN
#include <nsIGenericFactory.h>
#include <nsCOMPtr.h>

#include "igeckotip.h"

static HWND caretWnd;
static POINT p;
static HWINEVENTHOOK hook;
static HMODULE hModule;
static HMODULE hm;
static int numScopes=1;
static int scopes[48];
static WCHAR* regexp;
static bool initcalled=false;
static bool mozCaretExists=false;
static bool winCaretExists=false;
static long scrX; static long scrY;
static RECT rc; static RECT drc; 
//static FILE* stream; // DEBUG

typedef int (__stdcall *SetInputScopes_t) 
	(
	HWND hwnd,  
	const int *pInputScopes,
	unsigned  cInputScopes,
	WCHAR  **ppszPhraseList,
	unsigned  cPhrases,
	WCHAR  *pszRegExp,
	WCHAR  *pszSRGS
	);

typedef int (__stdcall *SetInputScope_t) 
	(
	HWND hwnd,  
	int inputscope
	);

SetInputScope_t mSetInputScope = 0;
SetInputScopes_t mSetInputScopes = 0;
HBITMAP mBitmap;
VARIANT v = { VT_I4, CHILDID_SELF};


void initDLL()
{
	initcalled=true;
	hm = ::LoadLibrary(TEXT("msctf.dll"));
	if (hm)
	{
		mSetInputScope = (SetInputScope_t)GetProcAddress(hm,"SetInputScope");
		mSetInputScopes = (SetInputScopes_t)GetProcAddress(hm,"SetInputScopes");
	}
	else
	{
		mSetInputScope = 0;
		mSetInputScopes = 0;
	}
}

VOID CALLBACK WinEventProc(
  HWINEVENTHOOK hWinEventHook,
  DWORD event,
  HWND hwnd,
  LONG idObject,
  LONG idChild,
  DWORD dwEventThread,
  DWORD dwmsEventTime
)
{
	IAccessible* pacc;
	VARIANT chID, role;

	if (!initcalled) initDLL();
	if (!mozCaretExists) return; 

	// the mozilla caret has moved-- update the win32 caret as needed. 
	// this code also creates the win32 caret it doesn't exist but the 
	// mozilla caret exists according to the extension's JS calling 
	// geckotip.focused 
	if ((idObject==OBJID_CARET)&&(event==EVENT_OBJECT_LOCATIONCHANGE)) {
		if (S_OK!=AccessibleObjectFromEvent(hwnd,idObject,idChild,&pacc,&chID)) return;
		pacc->get_accRole(chID,&role);

		if (role.lVal == ROLE_SYSTEM_CARET)
		{
			caretWnd=hwnd;

			long w,h;
			pacc->accLocation(&p.x,&p.y,&w,&h,v);
			scrX=p.x; scrY=p.y;

			// Convert Screen to window coordinates
			ScreenToClient(hwnd,&p);

			if (!winCaretExists) {
				HWND hwndF = GetFocus();
			
				//winCaretExists = ::CreateCaret(hwnd,NULL,10,10); // Visible Carat for debugging
				winCaretExists = ::CreateCaret(hwnd,mBitmap,10,10);
				if (winCaretExists) {
					::ShowCaret(hwnd);
					if (hwndF) {
						if (mSetInputScopes(hwndF,&scopes[0],numScopes,NULL,0,regexp,NULL) != S_OK) {
							scopes[0]=0;
							mSetInputScopes(hwndF,&scopes[0],1,NULL,0,NULL,NULL);
						}
					}
				}
			}
			if (winCaretExists) ::SetCaretPos(p.x,p.y);
		}
		pacc->Release();	
	}
}


BOOL APIENTRY DllMain( HMODULE hModule1, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	switch (ul_reason_for_call)
	{
	case DLL_PROCESS_ATTACH:
		{			

		int d=0;
		mBitmap = CreateBitmap(1,1,1,1,&d);
		hModule=hModule1;

		//stream = fopen("outfile.txt","w"); // DEBUG


		hook = SetWinEventHook(
			//EVENT_SYSTEM_MOVESIZESTART, 
			EVENT_OBJECT_LOCATIONCHANGE, // EVENT_MIN
			EVENT_OBJECT_LOCATIONCHANGE, // EVENT_MAX,
			hModule,
			WinEventProc,
			GetCurrentProcessId(),
			0,
			WINEVENT_INCONTEXT
			);
		}		
		break;
	case DLL_PROCESS_DETACH:
		UnhookWinEvent(hook);
		FreeLibrary(hm);
		if (regexp != NULL) {
			free(regexp);
			regexp=NULL;
		}
		break;
	}
    return TRUE;
}

class dgGeckoTip : public iGeckotip
{
public:
  NS_DECL_ISUPPORTS;
  NS_DECL_IGECKOTIP;

  dgGeckoTip(){}

private:

	~dgGeckoTip(){}

};

NS_IMETHODIMP dgGeckoTip::Noop()
{
	return NS_OK;
}

NS_IMETHODIMP dgGeckoTip::Unfocused()
{
	HWND hwndF;

	hwndF=GetFocus();
	mozCaretExists=false;
	winCaretExists=false;
	DestroyCaret();
	return NS_OK;
}

NS_IMETHODIMP dgGeckoTip::Focused() 
{
	mozCaretExists=true;
	return NS_OK;
}


NS_IMETHODIMP dgGeckoTip::SetGrabPanel(PRInt32 p)
{
	HWND hwndP;
	HWND hwndF = GetFocus();
	long panX=scrX; long panY=scrY;

	hwndP = FindWindow("com.ibm.sharkapp.SharkWindow", NULL);
	if (hwndP==NULL) hwndP = FindWindow(NULL, "MessagEase");
	if (hwndP==NULL) hwndP = FindWindow(NULL, "AlphaTapMain");
	if (hwndP==NULL) hwndP = FindWindow(NULL, "Fitaly");
	if (hwndP==NULL) hwndP = FindWindow(NULL, "TAPLESSxp");
	if (hwndP==NULL) hwndP = FindWindow("WordLogicWndCls", NULL);
	if (p > 0) {
		if (hwndP!=NULL) {
			ShowWindow(hwndP,SW_SHOWNOACTIVATE);
			SetForegroundWindow(hwndF);
			// Make sure the window will fit on the screen
			GetWindowRect(GetDesktopWindow(), &drc);
			GetWindowRect(hwndP, &rc);
			if (winCaretExists) {
				if (p==1) {
					if (panX+rc.right-rc.left > drc.right)
						panX=drc.right-rc.right+rc.left;
				} 
				if (p==2) {
					panX=panX-(rc.right-rc.left);
					if (panX<0) panX=0;
				}
				if (panY+40+rc.bottom-rc.top > drc.bottom) {
					SetWindowPos(hwndP, HWND_TOP, panX, panY-40-(rc.bottom-rc.top), 0, 0, SWP_NOSIZE);
				} else {
					SetWindowPos(hwndP, HWND_TOP, panX, panY+40, 0, 0, SWP_NOSIZE);
				}
				SetForegroundWindow(hwndF);
			}
		}
	} else {
		if (hwndP!=NULL) {
			if (IsWindowVisible(hwndP)) {
				ShowWindow(hwndP,SW_HIDE); 
				SetForegroundWindow(hwndF);
			}
		}
	}
	return NS_OK;
}

NS_IMETHODIMP dgGeckoTip::ToggleDebug()
{
	return NS_OK;
}

NS_IMETHODIMP dgGeckoTip::SetScope(PRInt32 s)
{
	scopes[0]=s; 
	numScopes=1;
	if (regexp != NULL) {
		free(regexp);
		regexp=NULL;
	}
	return NS_OK;
}

NS_IMETHODIMP dgGeckoTip::SetRegExp(const PRUnichar *re)
{
	numScopes=0;
	if (regexp != NULL) {
		free(regexp);
		regexp=NULL;
	}
	regexp=_wcsdup(re);
    return NS_OK;
}

NS_IMETHODIMP dgGeckoTip::AddScope(PRInt32 s)
{
	if (numScopes < 47) {
		scopes[numScopes]=s; 
		numScopes++;
	}
	return NS_OK;
}

NS_IMETHODIMP dgGeckoTip::PutCaret(PRInt32 x, PRInt32 y)
{
	HWND hwndF = GetFocus();
	HWND hwndP = FindWindow("com.ibm.sharkapp.SharkWindow", NULL);
	if (hwndP==NULL) hwndP = FindWindow(NULL, "MessagEase");
	if (hwndP==NULL) hwndP = FindWindow(NULL, "AlphaTapMain");
	if (hwndP==NULL) hwndP = FindWindow(NULL, "Fitaly");
	if (hwndP==NULL) hwndP = FindWindow(NULL, "TAPLESSxp");
	if (hwndP==NULL) hwndP = FindWindow("WordLogicWndCls", NULL);
	if (hwndP!=NULL) {
		if (IsWindowVisible(hwndP)) {
			BringWindowToTop(hwndP);
			SetForegroundWindow(hwndF);
		}
	}
	return NS_OK;
}

NS_IMETHODIMP dgGeckoTip::NoteEvent(PRInt32 e)
{
	return NS_OK;
}

NS_IMPL_ISUPPORTS1(dgGeckoTip, iGeckotip) // nsISelectionListener,nsIDOMFocusListener)

////////////////////////////////////////////////////////////////
#define DG_GECKOTIP_CID { 0xce58b594, 0x4842, 0x11da, 0x95, 0x2, 0x0, 0xe0, 0x81, 0x61, 0x16, 0x5f }

NS_GENERIC_FACTORY_CONSTRUCTOR(dgGeckoTip);
static const nsModuleComponentInfo components[] =
{
	{ "GeckoTIP",
	DG_GECKOTIP_CID,
	"@geckotip.mozdev.org/geckotip/fx18;1",
	dgGeckoTipConstructor
	}
};
NS_IMPL_NSGETMODULE(nsGeckoTipModule, components)
